r"""

**openpnm.models.misc**

----

This submodule contains models for calculating general or generic values
that are useful for all other pore-scale models, such as scaling values, or
generating an array of random numbers.

"""

from .misc import constant
from .misc import product
from .misc import random
from .misc import scaled
from .misc import linear
from .misc import polynomial
from .misc import weibull
from .misc import normal
from .misc import generic
from .misc import from_neighbor_throats
from .misc import from_neighbor_pores
